import logging
import re

import httpx
from fastapi import HTTPException

from llm_bridge.client.implementations.claude.claude_response_handler import process_claude_non_stream_response
from llm_bridge.client.model_client.claude_client import ClaudeClient
from llm_bridge.type.chat_response import ChatResponse
from llm_bridge.type.serializer import serialize


class NonStreamClaudeClient(ClaudeClient):
    async def generate_non_stream_response(self) -> ChatResponse:
        try:
            logging.info(f"messages: {self.messages}")

            message = await self.client.beta.messages.create(
                model=self.model,
                max_tokens=self.max_tokens,
                temperature=self.temperature,
                system=self.system,
                messages=serialize(self.messages),
                betas=self.betas,
                tools=self.tools,
                thinking=self.thinking,
                extra_body=self.extra_body,
            )

            return await process_claude_non_stream_response(
                message=message,
                input_tokens=self.input_tokens,
                client=self.client,
                model=self.model,
            )
        except httpx.HTTPStatusError as e:
            status_code = e.response.status_code
            text = e.response.text
            raise HTTPException(status_code=status_code, detail=text)
        except Exception as e:
            logging.exception(e)
            match = re.search(r'\d{3}', str(e))
            if match:
                error_code = int(match.group(0))
            else:
                error_code = 500

            raise HTTPException(status_code=error_code, detail=str(e))



from typing import Any

from fastapi import HTTPException

from llm_bridge.client.chat_client import ChatClient
from llm_bridge.logic.chat_generate.model_client_factory.claude_client_factory import create_claude_client
from llm_bridge.logic.chat_generate.model_client_factory.gemini_client_factory import create_gemini_client
from llm_bridge.logic.chat_generate.model_client_factory.openai_client_factory import create_openai_client
from llm_bridge.type.message import Message


async def create_chat_client(
        api_keys: dict[str, str],
        messages: list[Message],
        model: str,
        api_type: str,
        temperature: float,
        stream: bool,
        thought: bool,
        web_search: bool,
        code_execution: bool,
        structured_output_schema: dict[str, Any] | None,
) -> ChatClient:
    if api_type == 'OpenAI':
        return await create_openai_client(
            api_keys={"OPENAI_API_KEY": api_keys["OPENAI_API_KEY"]},
            messages=messages,
            model=model,
            api_type=api_type,
            temperature=temperature,
            stream=stream,
            thought=thought,
            web_search=web_search,
            code_execution=code_execution,
            structured_output_schema=structured_output_schema,
        )
    elif api_type == 'OpenAI-Azure':
        return await create_openai_client(
            api_keys={
                "AZURE_API_KEY": api_keys["AZURE_API_KEY"],
                "AZURE_API_BASE": api_keys["AZURE_API_BASE"]
            },
            messages=messages,
            model=model,
            api_type=api_type,
            temperature=temperature,
            stream=stream,
            thought=thought,
            web_search=web_search,
            code_execution=code_execution,
            structured_output_schema=structured_output_schema,
        )
    elif api_type == 'OpenAI-GitHub':
        return await create_openai_client(
            api_keys={"GITHUB_API_KEY": api_keys["GITHUB_API_KEY"]},
            messages=messages,
            model=model,
            api_type=api_type,
            temperature=temperature,
            stream=stream,
            thought=thought,
            web_search=web_search,
            code_execution=code_execution,
            structured_output_schema=structured_output_schema,
        )
    elif api_type == 'Grok':
        return await create_openai_client(
            api_keys={"XAI_API_KEY": api_keys["XAI_API_KEY"]},
            messages=messages,
            model=model,
            api_type=api_type,
            temperature=temperature,
            stream=stream,
            thought=thought,
            web_search=web_search,
            code_execution=code_execution,
            structured_output_schema=structured_output_schema,
        )
    elif api_type == 'Google AI Studio Free Tier':
        return await create_gemini_client(
            api_key=api_keys["GOOGLE_AI_STUDIO_FREE_TIER_API_KEY"],
            vertexai=False,
            messages=messages,
            model=model,
            temperature=temperature,
            stream=stream,
            thought=thought,
            web_search=web_search,
            code_execution=code_execution,
            structured_output_schema=structured_output_schema,
        )
    elif api_type == 'Google AI Studio':
        return await create_gemini_client(
            api_key=api_keys["GOOGLE_AI_STUDIO_API_KEY"],
            vertexai=False,
            messages=messages,
            model=model,
            temperature=temperature,
            stream=stream,
            thought=thought,
            web_search=web_search,
            code_execution=code_execution,
            structured_output_schema=structured_output_schema,
        )
    elif api_type == 'Vertex AI':
        return await create_gemini_client(
            api_key=api_keys["VERTEX_AI_API_KEY"],
            vertexai=True,
            messages=messages,
            model=model,
            temperature=temperature,
            stream=stream,
            thought=thought,
            web_search=web_search,
            code_execution=code_execution,
            structured_output_schema=structured_output_schema,
        )
    elif api_type == 'Claude':
        return await create_claude_client(
            api_key=api_keys["ANTHROPIC_API_KEY"],
            messages=messages,
            model=model,
            temperature=temperature,
            stream=stream,
            thought=thought,
            web_search=web_search,
            code_execution=code_execution,
            structured_output_schema=structured_output_schema,
        )
    else:
        raise HTTPException(status_code=400, detail="Invalid API type")



import asyncio
import math
import os
import time
from pathlib import Path

from dotenv import load_dotenv

from llm_bridge import *
from usage.workflow import workflow

script_dir = Path(__file__).parent.resolve()
load_dotenv(script_dir / ".env")

Timeout_Seconds = 60
Trial_Count = 3

api_keys = {
    "OPENAI_API_KEY": os.environ.get("OPENAI_API_KEY"),
    "AZURE_API_KEY": os.environ.get("AZURE_API_KEY"),
    "AZURE_API_BASE": os.environ.get("AZURE_API_BASE"),
    "GITHUB_API_KEY": os.environ.get("GITHUB_API_KEY"),
    "GOOGLE_AI_STUDIO_FREE_TIER_API_KEY": os.environ.get("GOOGLE_AI_STUDIO_FREE_TIER_API_KEY"),
    "GOOGLE_AI_STUDIO_API_KEY": os.environ.get("GOOGLE_AI_STUDIO_API_KEY"),
    "VERTEX_AI_API_KEY": os.environ.get("VERTEX_AI_API_KEY"),
    "ANTHROPIC_API_KEY": os.environ.get("ANTHROPIC_API_KEY"),
    "XAI_API_KEY": os.environ.get("XAI_API_KEY"),
}


async def measure_first_chunk_latency(
        api_type: str,
        model: str,
) -> float:
    messages = [
        Message(
            role=Role.User,
            contents=[Content(type=ContentType.Text, data="Hello")]
        )
    ]

    start = time.perf_counter()
    response = await workflow(
        api_keys,
        messages,
        model,
        api_type,
        temperature=0,
        stream=True,
        thought=False,
        web_search=False,
        code_execution=False,
        structured_output_schema=None,
    )

    async for _ in response:
        return time.perf_counter() - start

    return time.perf_counter() - start


async def main():
    model_prices = get_model_prices()

    results: list[tuple[str, str, float | None, float | None, str]] = []

    for price in model_prices:
        api_type: str = price["apiType"]
        model: str = price["model"]

        latencies: list[float] = []
        errors: set[str] = set()

        for trial in range(Trial_Count):
            try:
                latency = await asyncio.wait_for(
                    measure_first_chunk_latency(api_type, model),
                    timeout=Timeout_Seconds,
                )
                latencies.append(latency)
                print(f"  [{trial + 1}/{Trial_Count}] OK      {api_type:20s} {model:35s} {latency:8.3f}s")
            except asyncio.TimeoutError:
                errors.add(f"TIMEOUT ({Timeout_Seconds}s)")
                print(f"  [{trial + 1}/{Trial_Count}] TIMEOUT {api_type:20s} {model:35s}")
            except Exception as e:
                errors.add(str(e))
                print(f"  [{trial + 1}/{Trial_Count}] ERROR   {api_type:20s} {model:35s} - {e}")

        avg_latency = sum(latencies) / len(latencies) if latencies else None
        std_latency = math.sqrt(sum((l - avg_latency) ** 2 for l in latencies) / len(latencies)) if len(latencies) == Trial_Count else None
        if errors:
            error_str = " | ".join(sorted(errors))
            status = f"ERRORS ({len(latencies)}/{Trial_Count} OK): {error_str}" if avg_latency is not None else f"FAILED: {error_str}"
        else:
            status = "OK"

        results.append((api_type, model, avg_latency, std_latency, status))
        avg_str = f"{avg_latency:.3f}s" if avg_latency is not None else "N/A"
        std_str = f"{std_latency:.3f}s" if std_latency is not None else "N/A"
        print(f"  => AVG  {api_type:20s} {model:35s} {avg_str:>8s} ± {std_str:<8s}  {status}\n")

    print(f"{'=' * 110}")
    print(f"{'API Type':20s} {'Model':35s} {'Avg Latency':>12s} {'Std Dev':>10s}  {'Status'}")
    print(f"{'-' * 110}")
    for api_type, model, avg_latency, std_latency, status in results:
        avg_str = f"{avg_latency:.3f}s" if avg_latency is not None else "N/A"
        std_str = f"{std_latency:.3f}s" if std_latency is not None else "N/A"
        print(f"{api_type:20s} {model:35s} {avg_str:>12s} {std_str:>10s}  {status}")
    print(f"{'=' * 110}")


if __name__ == "__main__":
    asyncio.run(main())



from llm_bridge.logic.model_prices import get_model_prices


def test_get_model_prices_types():
    result = get_model_prices()
    for model_price in result:
        assert isinstance(model_price["apiType"], str)
        assert isinstance(model_price["model"], str)
        assert isinstance(model_price["input"], float)
        assert isinstance(model_price["output"], float)



from typing import Any

from google import genai
from google.genai import types
from google.genai.types import Modality, MediaResolution

from llm_bridge.client.implementations.gemini.non_stream_gemini_client import NonStreamGeminiClient
from llm_bridge.client.implementations.gemini.stream_gemini_client import StreamGeminiClient
from llm_bridge.logic.chat_generate.chat_message_converter import convert_messages_to_gemini
from llm_bridge.logic.message_preprocess.message_preprocessor import extract_system_messages
from llm_bridge.type.message import Message


async def create_gemini_client(
        api_key: str,
        vertexai: bool,
        messages: list[Message],
        model: str,
        temperature: float,
        stream: bool,
        thought: bool,
        web_search: bool,
        code_execution: bool,
        structured_output_schema: dict[str, Any] | None,
):
    client = genai.Client(
        vertexai=vertexai,
        api_key=api_key,
    )

    system_instruction = await extract_system_messages(messages) or " "
    tools = []
    thinking_config = None
    response_modalities = [Modality.TEXT]

    if web_search:
        tools.append(
            types.Tool(
                google_search=types.GoogleSearch()
            )
        )
    if thought:
        thinking_config = types.ThinkingConfig(
            include_thoughts=True,
            thinking_budget=-1,
        )
    if "image" not in model:
        tools.append(
            types.Tool(
                url_context=types.UrlContext()
            )
        )
        if not vertexai:
            if code_execution:
                tools.append(
                    types.Tool(
                        code_execution=types.ToolCodeExecution()
                    )
                )
    if "image" in model:
        response_modalities = [Modality.TEXT, Modality.IMAGE]

    config = types.GenerateContentConfig(
        system_instruction=system_instruction,
        temperature=temperature,
        safety_settings=[
            types.SafetySetting(
                category=types.HarmCategory.HARM_CATEGORY_HATE_SPEECH,
                threshold=types.HarmBlockThreshold.OFF,
            ),
            types.SafetySetting(
                category=types.HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT,
                threshold=types.HarmBlockThreshold.OFF,
            ),
            types.SafetySetting(
                category=types.HarmCategory.HARM_CATEGORY_HARASSMENT,
                threshold=types.HarmBlockThreshold.OFF,
            ),
            types.SafetySetting(
                category=types.HarmCategory.HARM_CATEGORY_SEXUALLY_EXPLICIT,
                threshold=types.HarmBlockThreshold.OFF,
            ),
            types.SafetySetting(
                category=types.HarmCategory.HARM_CATEGORY_CIVIC_INTEGRITY,
                threshold=types.HarmBlockThreshold.OFF,
            ),
        ],
        tools=tools,
        thinking_config=thinking_config,
        response_modalities=response_modalities,
    )

    if vertexai:
        config.media_resolution=MediaResolution.MEDIA_RESOLUTION_HIGH

    if structured_output_schema:
        config.response_mime_type = "application/json"
        config.response_json_schema = structured_output_schema

    gemini_messages = await convert_messages_to_gemini(messages)

    if stream:
        return StreamGeminiClient(
            model=model,
            messages=gemini_messages,
            temperature=temperature,
            client=client,
            config=config,
        )
    else:
        return NonStreamGeminiClient(
            model=model,
            messages=gemini_messages,
            temperature=temperature,
            client=client,
            config=config,
        )



from google.genai import types


async def count_gemini_tokens(
        response: types.GenerateContentResponse
) -> tuple[int, int]:
    usage_metadata = response.usage_metadata
    if usage_metadata is None:
        return 0, 0
    input_tokens = usage_metadata.prompt_token_count
    if input_tokens is None: # For Vertex AI
        input_tokens = 0
    output_tokens = usage_metadata.candidates_token_count
    if output_tokens is None:
        total_token_count = usage_metadata.total_token_count
        if total_token_count is None: # For Vertex AI
            total_token_count = 0
        output_tokens = total_token_count - input_tokens
    return input_tokens, output_tokens




from typing import AsyncGenerator, Iterable

import openai.lib.azure
from openai.types import Reasoning
from openai.types.responses import ToolParam, ResponseIncludable, ResponseTextConfigParam

from llm_bridge.client.chat_client import ChatClient
from llm_bridge.type.chat_response import ChatResponse
from llm_bridge.type.model_message.openai_message import OpenAIMessage
from llm_bridge.type.model_message.openai_responses_message import OpenAIResponsesMessage


class OpenAIClient(ChatClient):
    def __init__(
            self,
            model: str,
            messages: list[OpenAIMessage | OpenAIResponsesMessage],
            temperature: float,
            api_type: str,
            client: openai.AsyncOpenAI | openai.lib.azure.AsyncAzureOpenAI,
            tools: Iterable[ToolParam] | None = None,
            reasoning: Reasoning | None = None,
            include: list[ResponseIncludable] | None = None,
            text: ResponseTextConfigParam | None = None,
    ):
        self.model = model
        self.messages = messages
        self.temperature = temperature
        self.api_type = api_type
        self.client = client
        self.tools = tools
        self.reasoning = reasoning
        self.include = include
        self.text = text

    async def generate_non_stream_response(self) -> ChatResponse:
        raise NotImplementedError

    async def generate_stream_response(self) -> AsyncGenerator[ChatResponse, None]:
        raise NotImplementedError
